/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.internal;

import org.diirt.datasource.DataSourceTypeAdapter;
import org.diirt.datasource.ValueCache;
import org.diirt.vtype.Array;
import org.diirt.vtype.VType;

import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for the type adapters used by the {@link JackieChannelHandler}.
 * 
 * @author Sebastian Marsching
 *
 * @param <VTypeType>
 *            type of the {@link VType} to which the child class extending this
 *            class converts.
 */
public abstract class JackieDataSourceTypeAdapter<VTypeType extends VType>
        implements
        DataSourceTypeAdapter<JackieConnectionPayload, JackieMessagePayload> {

    private final boolean needsControlsValue;
    private final ChannelAccessValueType supportedNativeType;
    private final Class<VTypeType> vTypeClass;
    private final boolean vTypeIsArray;

    /**
     * Initializes this type adapter with the specified values.
     * 
     * @param needsControlsValue
     *            <code>true</code> if this type adapter needs both a controls
     *            and time value, <code>false</code> if a time value is
     *            sufficient.
     * @param supportedNativeType
     *            native Channel Access type for which this adapter can convert
     *            values. This must be a type for which
     *            {@code supportedNativeType.isSimpleType()} returns
     *            <code>true</code>.
     * @param vTypeClass
     *            type to which this type adapter converts.
     */
    public JackieDataSourceTypeAdapter(boolean needsControlsValue,
            ChannelAccessValueType supportedNativeType,
            Class<VTypeType> vTypeClass) {
        // The constructor is only used internally, so we use assertions instead
        // real checks.
        assert (supportedNativeType != null);
        assert (supportedNativeType.isSimpleType());
        assert (vTypeClass != null);
        this.needsControlsValue = needsControlsValue;
        this.supportedNativeType = supportedNativeType;
        this.vTypeClass = vTypeClass;
        this.vTypeIsArray = Array.class.isAssignableFrom(vTypeClass);
    }

    @Override
    public Object getSubscriptionParameter(ValueCache<?> cache,
            JackieConnectionPayload connection) {
        // This method should never be called, so we do not implement it.
        throw new UnsupportedOperationException(
                "getSubscriptionParameter is not implemented by this type adapter.");
    }

    @Override
    public int match(ValueCache<?> cache, JackieConnectionPayload connection) {
        if (cache.getType().isAssignableFrom(vTypeClass)
                && supportedNativeType == connection.getNativeDataType()) {
            // If the target type represents an array, we can assign both arrays
            // and scalars. However, we prefer to map scalars to scalars, so we
            // give the adapter a lower priority.
            // If the target type represents a scalar, we need exactly one
            // element, so we cannot use the adapter if the element count is not
            // one.
            // We use priorities, so that specific type adapters (e.g. for
            // strings transmitted as waveforms) can override a generic type
            // adapter.
            // At the moment specifying different priorities (in the return
            // value) does not have any effect, but this might change in the
            // future.
            if (vTypeIsArray) {
                if (connection.getNativeCount() == 1) {
                    return 3;
                } else {
                    return 2;
                }
            } else {
                if (connection.getNativeCount() == 1) {
                    return 2;
                } else {
                    return 0;
                }
            }
        } else {
            return 0;
        }
    }

    @Override
    public boolean updateCache(ValueCache<?> cache,
            JackieConnectionPayload connection, JackieMessagePayload message) {
        // If there is an exception and the last time this method got called we
        // did not have an exception, we throw the exception now. This way, the
        // subscriptions get notified when there is a problem, but repeated
        // exceptions do not trigger the subscriptions again and again.
        if (message.hasException()) {
            if (message.isRepeatedException()) {
                return false;
            } else {
                Throwable e = message.getLastException();
                if (e instanceof RuntimeException) {
                    throw (RuntimeException) e;
                } else {
                    throw new RuntimeException(e.getMessage(), e);
                }
            }
        }
        // If the time value is null, we return. If the controls value is null
        // and we need it, we also return.
        if (message.getTimeValue() == null
                || (needsControlsValue && message.getControlsValue() == null)) {
            return false;
        }
        // The type of the value stored in the message should always match the
        // supportedNativeType. Other values should already have been discarded
        // in the event listener, so we only use an assertion here.
        assert (message.getControlsValue() == null || message.getControlsValue()
                .getType().toSimpleType() == supportedNativeType);
        assert (message.getTimeValue().getType()
                .toSimpleType() == supportedNativeType);
        VTypeType newValue = convertValue(connection, message);
        // We checked in match(...) that the value cache accepts our VType, so
        // we can safely do the cast here.
        @SuppressWarnings("unchecked")
        ValueCache<? super VTypeType> typedValueCache = (ValueCache<? super VTypeType>) cache;
        typedValueCache.writeValue(newValue);
        return true;
    }

    /**
     * Converts the ChannelAccess value to the target type. This method is
     * called by
     * {@link #updateCache(ValueCache, JackieConnectionPayload, JackieMessagePayload)}
     * after checking that the connection and message payloads actually contain
     * the required data for doing the conversion. Therefore, this method only
     * has to implement the type-specific logic.
     * 
     * @param connection
     *            connection payload for the current connection state.
     * @param message
     *            message payload storing the value that shall be converted.
     * @return converted value.
     */
    protected abstract VTypeType convertValue(
            JackieConnectionPayload connection, JackieMessagePayload message);

}
