/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.util;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;

/**
 * <p>
 * Simple JSON parser. This parser is optimized for simplicity, not performance,
 * so code that wants to parse large or complex JSON documents should use a
 * different JSON parser.
 * </p>
 * 
 * <p>
 * This parser has specifically been written in order to minimize the
 * dependencies needed for parsing JSON document. It only uses the Java 8 SE API
 * and the Apache Commons Lang 3 library.
 * </p>
 * 
 * <p>
 * This parser is able to parse any document that complies with the JSON
 * (ECMA-404) standard. Compared to many other parsers, this parser is very
 * strict about compliance and will typically refuse any input that is not
 * strictly compliant.
 * </p>
 * 
 * <p>
 * This parser converts JSON objects to Java objects using the following rules:
 * </p>
 * 
 * <ul>
 * <li>A JSON object is converted to a {@link Map Map&lt;String, Object&gt;}.
 * The order of the members is preserved in the map. The parser does not allow
 * duplicate member keys in objects. If a member using the same key as an
 * earlier member is found, the parser throws an exception.</li>
 * <li>A JSON array is converted to a {@link List List&lt;Object&gt;}.</li>
 * <li>A JSON string is converted to a {@link String}.</li>
 * <li>A JSON number is converted to a {@link Number}. The actual type of the
 * {@code Number} depends on the number's value and should be regarded as an
 * implementation detail that might change in the future.</li>
 * <li>A JSON boolean value is converted to a {@link Boolean}.</li>
 * <li>A JSON value of <code>null</code> is converted to <code>null</code>.</li>
 * </ul>
 * 
 * @see SimpleJsonGenerator
 * 
 * @author Sebastian Marsching
 */
public class SimpleJsonParser {

    /**
     * Parses the specified string into a Java object. Please refer to the
     * {@linkplain SimpleJsonParser class description} for details.
     * 
     * @param jsonString
     *            string that represents a valid JSON document.
     * @return object that is the result of converting the string from JSON into
     *         a Java object. <code>null</code> if and only if the
     *         <code>jsonString</code> is the literal string "null".
     * @throws IllegalArgumentException
     *             if the <code>jsonString</code> cannot be parsed because it is
     *             either invalid, or there is an object with duplicate member
     *             keys.
     */
    public static Object parse(String jsonString) {
        // If jsonString is null, fail early.
        if (jsonString == null) {
            throw new NullPointerException();
        }
        return new SimpleJsonParser(jsonString).parse();
    }

    private String parsedString;
    private int position;

    private SimpleJsonParser(String jsonString) {
        this.parsedString = jsonString;
        this.position = 0;
    }

    private boolean accept(int codePoint) {
        if (isNext(codePoint)) {
            consumeCodePoint();
            return true;
        } else {
            return false;
        }
    }

    private boolean accept(String acceptedString) {
        if (StringUtils.startsWith(parsedString.substring(position),
                acceptedString)) {
            position += acceptedString.length();
            return true;
        } else {
            return false;
        }
    }

    private Optional<Integer> acceptAnyOf(String options) {
        if (exhausted()) {
            return Optional.empty();
        }
        int actualCodePoint = peek();
        for (int expectedCodePoint, position = 0; position < options
                .length(); position += Character.charCount(expectedCodePoint)) {
            expectedCodePoint = options.codePointAt(position);
            if (actualCodePoint == expectedCodePoint) {
                consumeCodePoint();
                return Optional.of(expectedCodePoint);
            }
        }
        return Optional.empty();
    }

    private void acceptWhitespace() {
        boolean isWhitespace = true;
        while (!exhausted() && isWhitespace) {
            int codePoint = peek();
            switch (codePoint) {
            case '\t':
            case '\n':
            case '\r':
            case ' ':
                consumeCodePoint();
                break;
            default:
                isWhitespace = false;
                break;
            }
        }
    }

    private int consumeCodePoint() {
        // We assume that this method is only called after checking that we have
        // not reached the end of the string.
        int codePoint = parsedString.codePointAt(position);
        position += Character.charCount(codePoint);
        return codePoint;
    }

    private String escapeAndShorten() {
        return escapeAndShorten(parsedString.substring(position));
    }

    private String escapeAndShorten(CharSequence cs) {
        int maxLength = 12;
        if (cs.length() < maxLength) {
            return StringEscapeUtils.escapeJava(cs.toString());
        } else {
            return StringEscapeUtils
                    .escapeJava(cs.subSequence(0, maxLength - 3) + "...");
        }
    }

    private boolean exhausted() {
        return position >= parsedString.length();
    }

    private void expect(int expectedCodePoint) {
        if (exhausted()) {
            throw new IllegalArgumentException("Expected '"
                    + new String(Character.toChars(expectedCodePoint))
                    + "', but found end-of-string.");
        }
        int actualCodePoint = consumeCodePoint();
        if (actualCodePoint != expectedCodePoint) {
            throw new IllegalArgumentException("Expected '"
                    + new String(Character.toChars(expectedCodePoint))
                    + "', but found '"
                    + new String(Character.toChars(actualCodePoint)) + "'.");
        }
    }

    private int expectAny(String description) {
        if (exhausted()) {
            throw new IllegalArgumentException(
                    "Expected " + description + ", but found end-of-string.");
        }
        int codePoint = peek();
        if (!Character.isValidCodePoint(codePoint)) {
            throw new IllegalArgumentException(
                    "Expected " + description
                            + ", but found invalid code point \\u" + StringUtils
                                    .leftPad(Integer.toString(codePoint, 16), 4)
                            + ".");
        }
        consumeCodePoint();
        return codePoint;
    }

    private int expectAnyOf(String options, String description) {
        if (exhausted()) {
            throw new IllegalArgumentException(
                    "Expected " + description + ", but found end-of-string.");
        }
        int actualCodePoint = peek();
        for (int expectedCodePoint, position = 0; position < options
                .length(); position += Character.charCount(expectedCodePoint)) {
            expectedCodePoint = options.codePointAt(position);
            if (actualCodePoint == expectedCodePoint) {
                consumeCodePoint();
                return expectedCodePoint;
            }
        }
        throw new IllegalArgumentException("Expected " + description
                + ", but found '"
                + new String(Character.toChars(actualCodePoint)) + "'.");
    }

    private int expectDecimalDigit() {
        return expectAnyOf("0123456789",
                "'0', '1', '2', '3', '4', '5', '6', '7', or '9'");
    }

    private int fourHexDigits() {
        StringBuilder fourDigits = new StringBuilder(4);
        while (fourDigits.length() < 4) {
            fourDigits.appendCodePoint(
                    expectAnyOf("0123456789ABCDEFabcdef", "hexadecimal digit"));
        }
        return Integer.valueOf(fourDigits.toString(), 16);
    }

    private boolean isNext(int codePoint) {
        if (!exhausted() && peek() == codePoint) {
            return true;
        } else {
            return false;
        }
    }

    private boolean isNextAnyOf(String options) {
        if (exhausted()) {
            return false;
        }
        int actualCodePoint = peek();
        for (int expectedCodePoint, position = 0; position < options
                .length(); position += Character.charCount(expectedCodePoint)) {
            expectedCodePoint = options.codePointAt(position);
            if (actualCodePoint == expectedCodePoint) {
                return true;
            }
        }
        return false;
    }

    private List<Object> jsonArray() {
        // We use an ArrayList because in general, it performs better than a
        // LinkedList.
        expect('[');
        acceptWhitespace();
        if (accept(']')) {
            return Collections.emptyList();
        }
        ArrayList<Object> members = new ArrayList<Object>();
        boolean arrayClosed = false;
        while (!arrayClosed) {
            members.add(jsonValue());
            acceptWhitespace();
            if (accept(']')) {
                arrayClosed = true;
            } else {
                expect(',');
                acceptWhitespace();
            }
        }
        return members;
    }

    private Number jsonNumber() {
        // First, we copy the number into a string builder. This way, we know
        // that we have a valid number and we know where it ends.
        StringBuilder sb = new StringBuilder();
        sb.append(jsonNumberIntPart());
        if (accept('.')) {
            sb.appendCodePoint('.');
            sb.append(jsonNumberDigitsPart(false));
        }
        Optional<Integer> eCodePoint = acceptAnyOf("eE");
        if (eCodePoint.isPresent()) {
            sb.appendCodePoint(eCodePoint.get());
            if (accept('+')) {
                sb.appendCodePoint('+');
            } else if (accept('-')) {
                sb.appendCodePoint('-');
            }
            sb.append(jsonNumberDigitsPart(false));
        }
        BigDecimal number = new BigDecimal(sb.toString());
        try {
            return number.byteValueExact();
        } catch (ArithmeticException e) {
            // Ignore any exception that might occur here, we simply continue
            // with other conversions.
        }
        try {
            return number.shortValueExact();
        } catch (ArithmeticException e) {
            // Ignore any exception that might occur here, we simply continue
            // with other conversions.
        }
        try {
            return number.intValueExact();
        } catch (ArithmeticException e) {
            // Ignore any exception that might occur here, we simply continue
            // with other conversions.
        }
        try {
            return number.longValueExact();
        } catch (ArithmeticException e) {
            // Ignore any exception that might occur here, we simply continue
            // with other conversions.
        }
        float numberAsFloat = number.floatValue();
        if (Float.isFinite(numberAsFloat)
                && BigDecimal.valueOf(numberAsFloat).equals(number)) {
            return numberAsFloat;
        }
        double numberAsDouble = number.doubleValue();
        if (Double.isFinite(numberAsDouble)
                && BigDecimal.valueOf(numberAsDouble).equals(number)) {
            return numberAsDouble;
        }
        try {
            return number.toBigIntegerExact();
        } catch (ArithmeticException e) {
            // Ignore any exception that might occur here, we simply return the
            // BigDecimal.
        }
        return number;
    }

    private CharSequence jsonNumberDigitsPart(boolean optional) {
        StringBuilder sb = new StringBuilder();
        if (!optional) {
            int digitCodePoint = expectDecimalDigit();
            sb.appendCodePoint(digitCodePoint);
        }
        Optional<Integer> nextDigitCodePoint = acceptAnyOf("0123456789");
        while (nextDigitCodePoint.isPresent()) {
            sb.appendCodePoint(nextDigitCodePoint.get());
            nextDigitCodePoint = acceptAnyOf("0123456789");
        }
        return sb;
    }

    private CharSequence jsonNumberIntPart() {
        StringBuilder sb = new StringBuilder();
        if (accept('-')) {
            sb.appendCodePoint('-');
        }
        int digitCodePoint = expectDecimalDigit();
        sb.appendCodePoint(digitCodePoint);
        if (digitCodePoint == '0') {
            return sb;
        }
        sb.append(jsonNumberDigitsPart(true));
        return sb;
    }

    private Map<String, Object> jsonObject() {
        expect('{');
        acceptWhitespace();
        if (accept('}')) {
            return Collections.emptyMap();
        }
        // We use a linked hash-map so that the order of members is
        // preserved.
        LinkedHashMap<String, Object> members = new LinkedHashMap<String, Object>();
        boolean objectClosed = false;
        while (!objectClosed) {
            Pair<String, Object> member = jsonObjectMember();
            // This is a SIMPLE parser, so we do not support duplicate keys
            // (even though the JSON specification basically allows them).
            if (members.put(member.getLeft(), member.getRight()) != null) {
                throw new IllegalArgumentException("Found duplicate key \""
                        + escapeAndShorten(member.getLeft()) + "\" in object.");
            }
            acceptWhitespace();
            if (accept('}')) {
                objectClosed = true;
            } else {
                expect(',');
                acceptWhitespace();
            }
        }
        return members;
    }

    private Pair<String, Object> jsonObjectMember() {
        String key = jsonString();
        acceptWhitespace();
        expect(':');
        acceptWhitespace();
        Object value = jsonValue();
        return Pair.of(key, value);
    }

    private String jsonString() {
        expect('"');
        StringBuilder content = new StringBuilder();
        boolean stringClosed = false;
        while (!stringClosed) {
            if (accept('"')) {
                stringClosed = true;
            } else if (accept('\\')) {
                int codePoint = expectAnyOf("\"\\/bfnrtu",
                        "any of '\"', '\\', '/', 'b', 'f', 'n', 'r', 't', or 'u'");
                switch (codePoint) {
                case '"':
                case '\\':
                case '/':
                    content.appendCodePoint(codePoint);
                    break;
                case 'b':
                    content.appendCodePoint('\b');
                    break;
                case 'f':
                    content.appendCodePoint('\f');
                    break;
                case 'n':
                    content.appendCodePoint('\n');
                    break;
                case 'r':
                    content.appendCodePoint('\r');
                    break;
                case 't':
                    content.appendCodePoint('\t');
                    break;
                case 'u':
                    // Unicode sequence.
                    int hexCodePoint = fourHexDigits();
                    if (!Character.isValidCodePoint(hexCodePoint)) {
                        throw new IllegalArgumentException(
                                "Illegal code point specified in unicode sequence \\u"
                                        + StringUtils.leftPad(Integer
                                                .toString(hexCodePoint, 16), 4)
                                        + ".");
                    }
                    content.appendCodePoint(hexCodePoint);
                    break;
                default:
                    // We matched all characters that we passed to the expect
                    // method, so we really should not find any other ones.
                    throw new RuntimeException("Internal logic error.");
                }
            } else {
                int codePoint = expectAny("valid string content");
                if (codePoint > 0 && codePoint < 0x20) {
                    throw new IllegalArgumentException(
                            "Expected valid string content, but found invalid control character 0x"
                                    + StringUtils.leftPad(
                                            Integer.toString(codePoint), 2, '0')
                                    + ".");
                }
                content.appendCodePoint(codePoint);
            }
        }
        return content.toString();
    }

    private Object jsonValue() {
        if (isNext('"')) {
            return jsonString();
        } else if (isNext('{')) {
            return jsonObject();
        } else if (isNext('[')) {
            return jsonArray();
        } else if (accept("true")) {
            return Boolean.TRUE;
        } else if (accept("false")) {
            return Boolean.FALSE;
        } else if (accept("null")) {
            return null;
        } else if (isNextAnyOf("-0123456789")) {
            return jsonNumber();
        } else {
            throw new IllegalArgumentException(
                    "Expected JSON value, but found \"" + escapeAndShorten()
                            + "\".");
        }
    }

    private Object parse() {
        // We always throw an IllegalArgumentException, so we can specifically
        // catch it.
        String errorMessage;
        try {
            Object obj = jsonValue();
            if (position < parsedString.length()) {
                throw new IllegalArgumentException(
                        "Expected end-of-string, but found \""
                                + escapeAndShorten() + "\".");
            }
            return obj;
        } catch (IllegalArgumentException e) {
            errorMessage = e.getMessage();
        }
        // We use the position information to determine where the problem
        // happened. This can help the user to find the problem in the document.
        int line = 0;
        int column = 0;
        int codePoint;
        boolean lastCharWasCr = false;
        for (int i = 0; i < position; i += Character.charCount(codePoint)) {
            codePoint = parsedString.codePointAt(i);
            // We ignore a newline directly after a carriage return, if we did
            // not, we would mess up our line count for documents using CR LF as
            // the end-of-line sequence.
            if (lastCharWasCr && codePoint == '\n') {
                lastCharWasCr = false;
                continue;
            }
            lastCharWasCr = false;
            if (codePoint == '\r') {
                lastCharWasCr = true;
                ++line;
                column = 0;
            } else if (codePoint == '\n') {
                ++line;
                column = 0;
            } else {
                ++column;
            }
        }
        // Most users expect one-based line and column numbers, so we add one
        // when including them in the error message.
        throw new IllegalArgumentException("Error at line " + (line + 1)
                + " column " + (column + 1) + ": " + errorMessage);
    }

    private int peek() {
        // We assume that this method is only called after checking that we have
        // not reached the end of the string.
        return parsedString.codePointAt(position);
    }
}
